import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Collections;
import java.util.Hashtable;

/**
 * Der Serverprozess fr den Spielserver.
 * 
 * @author Bernd Gramlich 
 * @version 2014-12-31
 */
public class ServerProzess implements Runnable
{
    /*
     * Der ServerSocket, auf dem ich laufe.
     */
    private ServerSocket serverSocket;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) den
     * zugehrigen Clientantwortprozess enthlt.
     */
    private Hashtable <String, ClientAntwortProzess> prozess;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) den
     * Nick der Wunschpartnerin fr das nchste Spiel enthlt.
     */
    private Hashtable <String, String> wunschpartnerin;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) sein
     * aktuelles Spiel enthlt.
     */
    private Hashtable <String, Spiel> spiel;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) seine
     * aktuelle Punktzahl enthlt.
     */
    private Hashtable <String, Integer> punkte;

    /**
     * Konstruktor.
     * Erzeuge einen Serverprozess auf dem angegebenen ServerSocket.
     * Beginne mit leeren Tabellen.
     */
    public ServerProzess (ServerSocket s)
    {
        serverSocket = s;

        prozess         = new Hashtable <String, ClientAntwortProzess> ();
        punkte          = new Hashtable <String, Integer> ();
        spiel           = new Hashtable <String, Spiel> ();
        wunschpartnerin = new Hashtable <String, String> ();
    }

    /**
     * Ablaufplan.
     * Warte auf neue Clientverbindungen. Starte dabei fr jeden
     * sich anmeldenden Client einen neuen ClientAntwortProzess.
     */
    public void run ()
    {
        try
        {
            while (true)
            {
                Socket clientSocket = serverSocket.accept ();
                new Thread (new ClientAntwortProzess (clientSocket, this)).start();
            }
        }
        catch (IOException fehler)
        {
            System.err.println ("Serverprozess scheitert.");
            System.exit (1);
        }
    }

    /**
     * Sende die angegebene Nachricht an alle.
     */
    public synchronized void sendeAnAlle (String nachricht)
    {
        for (ClientAntwortProzess c: prozess.values ())
        {
            c.sende (nachricht);
        }
    }

    /**
     * Sende die angegebene Nachricht an die angegebene Person.
     */
    public synchronized boolean sendeAn (String name, String nachricht, ClientAntwortProzess absender)
    {
        if (prozess.containsKey (name))
        {
            ClientAntwortProzess empfaenger = prozess.get (name);
            empfaenger.sende (nachricht);

            return true;
        }
        else
        {
            absender.sende ("There's no one named " + name + ".");

            return false;
        }
    }

    /**
     * Erstelle oder ndere den Nick eines Clientantwortprozesses.
     */
    public synchronized boolean aendereNick (String alterName, String neuerName, ClientAntwortProzess absender)
    {
        if (prozess.containsKey (neuerName))
        {
            absender.sende ("Somebody is already called " + neuerName + ".");
            
            return false;
        }
        else
        { 
            prozess.put (neuerName, absender);

            /*
             * Falls der Name schon existierte ...
             */
            if (alterName != null)
            {
                /*
                 * ... ndere seine Identitt in allen Tabellen.
                 */
                punkte.put (neuerName, punkte.get (alterName));

                if (spiel.containsKey (alterName))
                {
                    spiel.put (neuerName, spiel.get (alterName));
                }

                if (wunschpartnerin.containsKey (alterName))
                {
                    wunschpartnerin.put (neuerName, wunschpartnerin.get (alterName));
                }

                for (String nutzerin: Collections.list (wunschpartnerin.keys ()))
                {
                    if (wunschpartnerin.get (nutzerin).equals (alterName))
                    {
                        wunschpartnerin.remove (nutzerin);
                        wunschpartnerin.put (nutzerin, neuerName);
                    }
                }

                prozess.remove (alterName);
                punkte.remove (alterName);
                spiel.remove (alterName);
                wunschpartnerin.remove (alterName);

                /*
                 * Mache die Namensnderung bekannt.
                 */
                sendeAnAlle (alterName + " will henceforth be known as " + neuerName + ".");
            }
            else
            {
                /*
                 * Andernfalls nimm die neue Benutzerin auf.
                 */
                punkte.put (neuerName, 0);
                sendeAnAlle (neuerName + " has joined.");
            }

            return true;
        }
    }

    /**
     * Melde den angegebenen Nick und seinen Clientantwortprozess ab.
     */
    public synchronized void meldeAb (String name)
    {
        prozess.remove (name);
        punkte.remove (name);
        spiel.remove (name);
        wunschpartnerin.remove (name);

        sendeAnAlle (name + " has left.");
    }

    /**
     * Teile der Benutzerin name alle Nicks und ihre Herkunft mit,
     * deren Nick mit muster beginnt.
     */
    public synchronized void nenneNicks (String name, String muster)
    {
        for (String nutzerin: Collections.list (prozess.keys ()))
        {
            if (nutzerin.startsWith (muster))
            {
                prozess.get (name).sende (nutzerin + " is connected from " + prozess.get (nutzerin).gibAdresse () + ".");
            }
        }
    }

    /**
     * Vermerke, dass die Spielerin sich ein Spiel mit der Gegnerin wnscht.
     * Wenn der Wunsch auf Gegenseitigkeit beruht, starte ein neues Spiel.
     */
    public synchronized void wuenscheSpiel (String spielerin, String gegnerin)
    {
        /*
         * Wenn die Gegnerin existiert ...
         */
        if (prozess.containsKey (gegnerin))
        {
            /*
             * ... vermerke sie als Wunschpartnerin der Spielerin.
             */
            wunschpartnerin.put (spielerin, gegnerin);

            /*
             * Wenn die Spielerin auch Wunschpartnerin der Gegnerin ist ...
             */
            if (spielerin.equals (wunschpartnerin.get (gegnerin)))
            {
                /*
                 * ... lose die Farben aus und ...
                 */
                double zufall = Math.random ();
                String gelb;
                String rot;

                if (Math.random () < 0.5)
                {
                    gelb = spielerin;
                    rot  = gegnerin;
                }
                else
                {
                    gelb = gegnerin;
                    rot  = spielerin;
                }

                /*
                 * ... beginne ein neues Spiel.
                 */
                Spiel neuesSpiel = new Spiel (gelb, rot);

                spiel.put (spielerin, neuesSpiel);
                spiel.put (gegnerin, neuesSpiel);

                /*
                 * Versetze beide Kontrahentinnen in den Zustand SPIELEN.
                 */
                prozess.get (spielerin).spiele (gegnerin);
                prozess.get (gegnerin).spiele (spielerin);

                /*
                 * Kndige das Spiel serverweit an.
                 */
                sendeAnAlle ("New game between " + gelb + " (yellow) and " + rot + " (red).");

                /*
                 * Sende beiden Clients das Signal zum Spielstart.
                 */
                prozess.get (spielerin).sende ("Playing.");
                prozess.get (gegnerin).sende ("Playing.");
            }
            /*
             * Falls die Gegnerin bereits spielt ...
             */
            else if (spiel.containsKey (gegnerin))
            {
                /*
                 * ... melde dies der Spielerin.
                 */
                prozess.get (spielerin).sende (gegnerin + " is already playing a game.");
            }
        }
    }

    /**
     * Spiele im eigenen Spiel in der angegebenen Spalte.
     */
    public synchronized void spiele (String name, String spalte)
    {
        /*
         * Wenn die Spielerin gerade ein Spiel spielt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * Ermittle dieses Spiel und die zugehrige Gegnerin.
             */
            Spiel meinSpiel = spiel.get (name);
            String gegnerin = wunschpartnerin.get (name);
            String gewinnerin;
            int spaltennummer;
            Ergebnis spielergebnis;

            /*
             * Versuche, in der angegebenen Spalte zu spielen.
             */
            try
            {
                spaltennummer = Integer.parseInt (spalte);
                spielergebnis = meinSpiel.wirfEin (name, spaltennummer);

                /*
                 * Reagiere je nach Ergebnis des Spielzugs.
                 */
                switch (spielergebnis)
                {
                    /*
                     * Falls die Spielerin nicht dran war, melde dies.
                     */
                    case NICHT_DRAN:
                    prozess.get (name).sende ("It's not your turn.");
                    break;

                    /*
                     * Falls die Spaltennummer ungltig war, melde dies.
                     */
                    case UNGUELTIG:
                    prozess.get (name).sende (spalte + " is not a valid column number.");
                    break;

                    /*
                     * Falls die Spalte voll war, melde dies.
                     */
                    case VOLL:
                    prozess.get (name).sende ("Column " + spalte + " is full.");
                    break;

                    /*
                     * Falls der Spielzug geklappt hat und das Spiel weiterluft,
                     * melde beiden Clients den Spielzug.
                     */
                    case WEITER:
                    prozess.get (name).sende ("moved " + spaltennummer);
                    prozess.get (gegnerin).sende ("moved " + spaltennummer);
                    break;

                    /*
                     * Falls das Spiel durch den Spielzug unentschieden endet ...
                     */
                    case UNENTSCHIEDEN:
                    /*
                     * ... melde den Spielzug.
                     */
                    prozess.get (name).sende ("moved " + spaltennummer);
                    prozess.get (gegnerin).sende ("moved " + spaltennummer);
                    /*
                     * Verknde und werte das Spiel als unentschieden.
                     */
                    sendeAnAlle (name + " drew against " + gegnerin + ".");
                    punkte.put (name, punkte.get (name) + 1);
                    punkte.put (gegnerin, punkte.get (gegnerin) + 1);
                    /*
                     * Sorge dafr, dass die Kontrahentinnen nicht mehr
                     * Wunschpartnerinnen voneinander sind (denn sonst knnte
                     * eine der beiden eine Revanche erzwingen).
                     */
                    wunschpartnerin.remove (name);
                    wunschpartnerin.remove (gegnerin);
                    /*
                     * Versetze beide Clients in den Zustand WERTEN.
                     */
                    prozess.get (name).werte ();
                    prozess.get (gegnerin).werte ();
                    break;

                    /*
                     * Falls das Spiel durch den Spielzug mit einem Sieg endet ...
                     */
                    case GELB: case ROT:
                    /*
                     * ... melde den Spielzug.
                     */
                    prozess.get (name).sende ("moved " + spaltennummer);
                    prozess.get (gegnerin).sende ("moved " + spaltennummer);

                    /*
                     * Verknde die Gewinnerin und gib ihr zwei Siegpunkte.
                     */
                    gewinnerin = meinSpiel.gibGewinnerin ();

                    if (gewinnerin.equals (name))
                    {
                        sendeAnAlle (name + " won against " + gegnerin + ".");
                        punkte.put (name, punkte.get (name) + 2);
                    }
                    else
                    {
                        sendeAnAlle (gegnerin + " won against " + name + ".");
                        punkte.put (gegnerin, punkte.get (gegnerin) + 2);
                    }

                    /*
                     * Sorge dafr, dass die Kontrahentinnen nicht mehr
                     * Wunschpartnerinnen voneinander sind (denn sonst knnte
                     * eine der beiden eine Revanche erzwingen).
                     */
                    wunschpartnerin.remove (name);
                    wunschpartnerin.remove (gegnerin);
                    /*
                     * Versetze beide Clients in den Zustand WERTEN.
                     */
                    prozess.get (name).werte ();
                    prozess.get (gegnerin).werte ();
                    break;
                }
            }
            /*
             * Falls die Spaltennummer keine Zahl war, melde dies.
             */
            catch (NumberFormatException fehler)
            {
                prozess.get (name).sende (spalte + " is not a column number.");
            }
        }
        /*
         * Falls die Spielerin kein Spiel am Laufen hat, melde dies.
         */
        else
        {
            prozess.get (name).sende ("You are not playing a game.");
        }
    }

    /**
     * Gib im eigenen Spiel auf.
     */
    public synchronized void gibAuf (String name)
    {
        /*
         * Falls die Spielerin gerade ein Spiel spielt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * ... ermittle ihre Gegnerin.
             */
            String gegnerin = wunschpartnerin.get (name);

            /*
             * Verknde die Niederlage und gib der Gegnerin zwei Siegpunkte.
             */
            sendeAnAlle (name + " resigns against " + gegnerin + ".");
            punkte.put (gegnerin, punkte.get (gegnerin) + 2);
            /*
             * Sorge dafr, dass die Kontrahentinnen nicht mehr
             * Wunschpartnerinnen voneinander sind (denn sonst knnte
             * eine der beiden eine Revanche erzwingen).
             */
            wunschpartnerin.remove (name);
            wunschpartnerin.remove (gegnerin);
            /*
             * Versetze beide Clients in den Zustand WERTEN.
             */
            prozess.get (name).werte ();
            prozess.get (gegnerin).werte ();
        }
        /*
         * Falls die Spielerin kein Spiel am Laufen hat, melde dies.
         */
        else
        {
            prozess.get (name).sende ("You are not playing a game.");
        }
    }

    /**
     * Beende das eigene Spiel.
     */
    public synchronized void beendeSpiel (String name)
    {
        /*
         * Falls die Spielerin noch ein Spiel besitzt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * ... lsche das Spiel.
             */
            spiel.remove (name);

            /*
             * Melde dem Client, dass das Spiel gelscht ist.
             */
            prozess.get (name).sende ("Done.");
            /*
             * Versetze den Client in den Zustand CHATTEN.
             */
            prozess.get (name).chatte ();
        }
        /*
         * Falls die Spielerin kein Spiel besitzt, melde dies.
         */
        else
        {
            prozess.get (name).sende ("You are not reviewing a game.");
        }
    }

    /**
     * Teile der Benutzerin name die Punkte aller Spielerinnen mit,
     * deren Nick mit muster beginnt.
     */
    public synchronized void nennePunkte (String name, String muster)
    {
        for (String nutzerin: Collections.list (punkte.keys ()))
        {
            if (nutzerin.startsWith (muster))
            {
                int punktzahl = punkte.get (nutzerin);
                
                if (1 == punktzahl)
                {
                    prozess.get (name).sende (nutzerin + " has " + punktzahl + " point.");
                }
                else
                {
                    prozess.get (name).sende (nutzerin + " has " + punktzahl + " points.");
                }
            }
        }
    }
}